//----------------------------------------------------------------------------

//      (USART0)

//----------------------------------------------------------------------------

/*
   :
         
   WAKE.     ,  UART
      :  
     .
      16- CRC (XMODEM),
     x^16 + x^12 + x^5 + 1 (0x1021).
  CRC   0xDE.
    UART :
     - 1200 
     - 8 
   - ODD
*/

//----------------------------------------------------------------------------

#include <Main.h>
#include <PortMod.h>
#include <Wake.h>
#include <Events.h>

//----------------------------- : -----------------------------------

#define MD_BAUD 1200   //   

#define MD_TIMEOUT 990 //     

#define UBRR0V (int)((FCLK * 1000000.0)/(16.0 * MD_BAUD) - 0.5) //

#define MD_RX_ADDR 0   //     
#define MD_TX_ADDR 0   //,   

#define MD_IDLE   0    //     IDLE
#define MD_DIR_TX 1    //    
#define MD_DIR_RX 2    //    
#define MD_INIT   3    //  

//------------------------------- : --------------------------------

char Md_Rx_Sta,        //   
     Md_Rx_Pre,        //  
     Md_Rx_Add,        //,    
     Md_Rx_Cmd,        // 
     Md_Rx_Nbt,        //    
     Md_Rx_Dat[FRAME], //  
     Md_Rx_Ptr;        //    
unsigned int Md_Rx_Crc; //   

volatile char Command; //   

char Md_Tx_Sta,        //   
     Md_Tx_Pre,        //  
     Md_Tx_Add,        //,   
     Md_Tx_Cmd,        //,   
     Md_Tx_Nbt,        //    
     Md_Tx_Dat[FRAME], //   
     Md_Tx_Ptr;        //    
unsigned int Md_Tx_Crc; //   

volatile int Md_Timer; // - 
char Md_Dir;           // 

//-------------------------------- : ----------------------------------

#pragma vector = USART0_TXC_vect
__interrupt void Md_Tx(void);     //   
#pragma vector = USART0_RXC_vect
__interrupt void Md_Rx(void);     //   
void Do_Crc16(char b, unsigned int *crc); //  
void Md_Set_Dir(char dir);        //  

//------------------------  UART: -------------------------------

void Md_Init_Port(void)
{
  //odd-parity, 8- 
  UCSR0C = (1<<UPM01) | (1<<UPM00) | (1<<UCSZ01) | (1<<UCSZ00);
  //  
  UBRR0L = LO(UBRR0V);
  UBRR0H = HI(UBRR0V);

  Md_Rx_Add = MD_RX_ADDR; //  
  Md_Tx_Add = MD_TX_ADDR; //  
  Md_Rx_Sta = WAIT_CARR;  // 
  Md_Tx_Sta = SEND_IDLE;  //   
  Command = CMD_NOP;      //   
  Md_Dir = MD_INIT;
  Md_Set_Dir(MD_IDLE);    //   
}

//--------------------   : ----------------------

void Md_Set_Dir(char dir)
{
  if(Md_Dir != dir)
  {
    switch(dir)
    {
    case MD_DIR_TX: //  
      {
        Port_M1_0;  //  !
        Port_M0_1;
        UCSR0B = (1<<TXCIE0) | (1<<TXEN0);
        Delay_m(26); //   DET  
        break;
      }
    case MD_DIR_RX: //  
      {
        Port_M0_0;
        Port_M1_1;
        while(UCSR0A & (1<<RXC0)) Md_Rx_Pre = UDR0; //  UART
        Command = CMD_NOP;
        UCSR0B = (1<<RXCIE0) | (1<<TXCIE0) | (1<<RXEN0) | (1<<TXEN0);
        Delay_m(1);  //
        break;
      }
    default:  //  MD_IDLE
      {
        Port_M0_0;
        Port_M1_1;
        UCSR0B = (1<<TXCIE0) | (1<<TXEN0);
        Delay_m(1);  //
      }
    }
    Md_Dir = dir;
  }
}

//------------------------  : ---------------------------

void Md_Detect(void)
{
  if(Md_Tx_Sta == SEND_OFF)
  {
    Md_Set_Dir(MD_IDLE);       //    IDLE
    Md_Tx_Sta = SEND_IDLE;     // 
    Md_Rx_Sta = WAIT_CARR;     //    
  }
  if(Md_Tx_Sta == SEND_IDLE)
  {
    if(Md_Rx_Sta == WAIT_CARR) //  
    {
      if(Pin_DET) //   
      {
        Md_Set_Dir(MD_DIR_RX); //  
        Md_Rx_Sta = WAIT_FEND; //  
      }
    }
    else //  
    {
      if(!Pin_DET) //  , 
      {
        Md_Set_Dir(MD_IDLE);   // 
        Md_Rx_Sta = WAIT_CARR; //    
      }
    }
  }
  else
  {
    if(!Md_Timer)
    {
      Md_Set_Dir(MD_IDLE);     //  
      Md_Tx_Sta = SEND_IDLE;
      Md_Rx_Sta = WAIT_CARR;   //    
      Event(EV_TX, ROOT, 0);   //DEBUG
    }
  }
}

//----------------------------  : ---------------------------------

#pragma vector = USART0_RXC_vect
__interrupt void Md_Rx(void)
{
  char error_flags = (UCSR0A & (1<<FE0)); //  
  char data_byte = UDR0;              // 
  __enable_interrupt();               //  

  if(Md_Rx_Sta == WAIT_OFF)           //----->   ,
    return;                           // 

  if(error_flags)                     //     
  {
    Md_Rx_Sta = WAIT_FEND;            //  
    Command = CMD_ERR;                //  
    return;
  }

  if(data_byte == FEND)               //   ,
  {
    Md_Rx_Pre = data_byte;            //  -,
    Md_Rx_Crc = CRC_INIT;             // CRC,
    Md_Rx_Sta = WAIT_ADDR;            //  ,
    Do_Crc16(data_byte, &Md_Rx_Crc);  // CRC,
    return;                           //
  }

  if(Md_Rx_Sta == WAIT_FEND)          //----->   FEND,
    return;                           // 

  char Pre = Md_Rx_Pre;               //  -
  Md_Rx_Pre = data_byte;              // -
  if(Pre == FESC)                     // -  FESC,
  {
    if(data_byte == TFESC)            //    TFESC,
      data_byte = FESC;               //    FESC
    else if(data_byte == TFEND)       //    TFEND,
           data_byte = FEND;          //    FEND
         else
         {
           Md_Rx_Sta = WAIT_FEND;     //     ,
           Command = CMD_ERR;         //  FESC, 
           return;
         }
  }
  else
  {
    if(data_byte == FESC)             //    FESC,  
      return;                         //  -
  }

  switch(Md_Rx_Sta)
  {
  case WAIT_ADDR:                     //----->   
    {
      if(data_byte & 0x80)            //  7    ,   
      {
        data_byte = data_byte & 0x7F; //  7,   
        if(data_byte == 0 || data_byte == Md_Rx_Add) //    ,
        {
          Do_Crc16(data_byte, &Md_Rx_Crc); //  CRC 
          Md_Rx_Sta = WAIT_CMD;       //   
          break;
        }
        Md_Rx_Sta = WAIT_FEND;        //  ,   
        break;
      }                               //  7   , 
      Md_Rx_Sta = WAIT_CMD;           //    
    }
  case WAIT_CMD:                      //----->   
    {
      if(data_byte & 0x80)            //  7 
      {
        Md_Rx_Sta = WAIT_FEND;        //  7   ,
        Command = CMD_ERR;            // 
        break;
      }
      Md_Rx_Cmd = data_byte;          // 
      Do_Crc16(data_byte, &Md_Rx_Crc); // CRC
      Md_Rx_Sta = WAIT_NBT;           //    
      break;
    }
  case WAIT_NBT:                      //----->    
    {
      if(data_byte > FRAME)           //   > FRAME,
      {
        Md_Rx_Sta = WAIT_FEND;
        Command = CMD_ERR;            // 
        break;
      }
      Md_Rx_Nbt = data_byte;
      Do_Crc16(data_byte, &Md_Rx_Crc); // CRC
      Md_Rx_Ptr = 0;                  //  
      Md_Rx_Sta = WAIT_DATA;          //   
      break;
    }
  case WAIT_DATA:                     //----->   
    {
      if(Md_Rx_Ptr < Md_Rx_Nbt)       //    ,
      {
        Md_Rx_Dat[Md_Rx_Ptr++] = data_byte; //   ,
        Do_Crc16(data_byte, &Md_Rx_Crc);  // CRC
        break;
      }
      if(data_byte != LO(Md_Rx_Crc))  // CRC LOW
      {
        Md_Rx_Sta = WAIT_FEND;        // CRC  ,
        Command = CMD_ERR;            // 
        break;
      }
      Md_Rx_Sta = WAIT_CRC;           //   CRC
      break;
    }
  case WAIT_CRC:
    {
      if(data_byte != HI(Md_Rx_Crc))  // CRC HIGH
      {
        Md_Rx_Sta = WAIT_FEND;        // CRC  ,
        Command = CMD_ERR;            // 
        break;
      }
      Md_Rx_Sta = WAIT_OFF;           //  
      Command = Md_Rx_Cmd;            //   
      break;
    }
  }
}

//----------------------------  : ------------------------------

#pragma vector = USART0_TXC_vect
__interrupt void Md_Tx(void)
{
  __enable_interrupt();               //  
  char data_byte;

  if(Md_Tx_Pre == FEND)               //  ,
  {
    data_byte = TFEND;                // TFEND  FEND
    Md_Tx_Pre = data_byte;
    UDR0 = data_byte;
    return;
  }
  if(Md_Tx_Pre == FESC)               //  ,
  {
    data_byte = TFESC;                // TFESC  FESC
    Md_Tx_Pre = data_byte;
    UDR0 = data_byte;
    return;
  }

  switch(Md_Tx_Sta)
  {
  case SEND_ADDR:                     //----->  
    {
      if(Md_Tx_Add)                   //    ,
      {
        data_byte = Md_Tx_Add | 0x80; //   ( 7  )
        Md_Tx_Sta = SEND_CMD;
        break;
      }
      else
        Md_Tx_Sta = SEND_CMD;         //   
    }
  case SEND_CMD:                      //----->  
    {
      data_byte = Md_Tx_Cmd & 0x7F;
      Md_Tx_Sta = SEND_NBT;
      break;
    }
  case SEND_NBT:                      //----->   
    {
      data_byte = Md_Tx_Nbt;
      Md_Tx_Sta = SEND_DATA;
      Md_Tx_Ptr = 0;                  //    
      break;
    }
  case SEND_DATA:                     //----->  
    {
      if(Md_Tx_Ptr < Md_Tx_Nbt)
        data_byte = Md_Tx_Dat[Md_Tx_Ptr++];
      else
      {
        data_byte = LO(Md_Tx_Crc);    // CRC LOW
        Md_Tx_Sta = SEND_CRC;
      }
      break;
    }
  case SEND_CRC:
    {
      data_byte = HI(Md_Tx_Crc);      // CRC HIGH
      Md_Tx_Sta = SEND_END;
      break;
    }
  default:
    {
      Md_Tx_Sta = SEND_OFF;           //  
      return;
    }
  }

  if(Md_Tx_Sta != SEND_CRC)
    Do_Crc16(data_byte, &Md_Tx_Crc);  // CRC
  Md_Tx_Pre = data_byte;              // -
  if(data_byte == FEND || data_byte == FESC)
    data_byte = FESC;                 // FESC,   
  UDR0 = data_byte;
}

//---------------------   : ------------------------

void Do_Crc16(char b, unsigned int *crc)
{
  *crc = *crc ^ (b << 8);
  for (char i = 0; i < 8; i++)
  {
    if (*crc & 0x8000) *crc = (*crc << 1) ^ 0x1021;
     else *crc <<= 1;
  }
}

//---------------------    : -------------------------

void Md_Tx_Frame(void)
{
  char data_byte = FEND;
  Md_Tx_Crc = CRC_INIT;            // CRC,
  Do_Crc16(data_byte, &Md_Tx_Crc); // CRC
  Md_Tx_Sta = SEND_ADDR;
  Md_Tx_Pre = TFEND;
  Md_Set_Dir(MD_DIR_TX);           //   
  UDR0 = data_byte;
  Md_Timer = MD_TIMEOUT;           // 
}

//----------------------------------------------------------------------------
